# -*- coding: utf-8 -*-

# artetv.py
# This file is part of qarte-4
#    
# Author: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
# Copyright: 2011-2018 Vincent Vande Vyvre
# Licence: GPL3
# Home page: https://launchpad.net/qarte
#
# arte TV Guide main class

import os
import sys
import shutil
import time
import pickle
import json
import glob
import urllib.request
import urllib.parse
from html.parser import HTMLParser

import html
import re
import logging
lgg = logging.getLogger(__name__)

from datetime import datetime, date, timedelta
from threading import Thread
 
from PyQt5.QtGui import QFont, QTextCharFormat
from PyQt5.QtCore import QObject, QCoreApplication, pyqtSignal, QLocale, QSize

from data import ARTE_TV
from tvdownloadingconfig import TvDownloadingConfig
from loader import Loader
from gui.warning import WarningBox


class ArteTV(QObject):
    """Define the arte TV Guide server handler.

    """
    parsingFinished = pyqtSignal(int)
    LDESC = re.compile("(?<=<div class='hidden-xs desc-wrapper')"
                        "(.*?)(?=</div>)",
                        re.DOTALL)
    # FIXME Quick fix, maybe is better to use htmlparser
    ALTLDESC = re.compile('(?<=<div class="program-description-full">)'
                        '(.*?)(?=</div>)',
                        re.DOTALL)
    DESC = re.compile("(?<=<meta content=.<p>)"
                        "(.*?)(?=</p>. name='description'>)", 
                        re.DOTALL)
    # FIXME Quick fix, maybe is better to use htmlparser
    ALTDESC = re.compile('(?<=<p class="program-description-short">)'
                        '(.*?)(?=</p>)',
                        re.DOTALL)
    DATE = re.compile("(?<=^<meta content=)(.*?)(?=name='date'>)", re.MULTILINE)
    ANNO = re.compile('(?<=<span class="program-casting-label">Année :</span>'
        '<span class="program-casting-name">)(.*?)(?=</span></li>)', re.MULTILINE)
    ORIG = re.compile('(?<=<li><span class="program-casting-label">Pays :</span>'
        '<span class="program-casting-name">)(.*?)(?=</span></li>)', re.DOTALL)
    LENGHT = re.compile('(?<=<span class="duration">)(.*?)(?=</span>)', re.DOTALL)
    def __init__(self, core):
        super().__init__()
        self.core = core
        self.ui = core.ui
        self.lang = self.core.cfg['lang']
        self.fatal_error = False
        self.cache = core.workspace['tv_tmp']
        self.videos = []
        self.is_loading = False
        self.loading_aborted = False
        lang = QLocale.system().name()
        self.locale = QLocale(lang)
        self.ui.tv_basket.itemAdded.connect(self.on_item_added)
        self.loader = Loader()
        self.loader.loadingProgress.connect(self.display_progress)
        self.loader.loadingFinished.connect(self.on_downloading_finished)
        self.parsingFinished.connect(self.remove_old_videos)

    def config_parser(self):
        """Configure the main pages parser.

        """
        self.load_summaries_index()
        links = self.get_site_pages()
        self.parser = Parser(ARTE_TV.replace('fr', self.lang))
        Thread(target=self.run_parser).start()

    def run_parser(self):
        """Run the parser.

        This executed into a thread.
        """
        begin = time.time()
        videos = self.parser.videos
        self.videos = [TVItem(item) for item in videos]
        self.indexes = self.parser.indexes
        for i in range(10):
            vids, idxs = self.parser.get_next_page()
            self.sort_videos(vids, idxs)

        self.check_thumbnails()
        videos_count = len(self.videos)
        end = time.time() - begin
        lgg.info('Found %s videos in %s sec.' % (videos_count, end))
        self.parsingFinished.emit(videos_count)

    def sort_videos(self, news, idxs):
        """Create the sorted (youngest to oldest) list of the videos.

        """
        lgg.info('Sorting videos')
        while news:
            item = news.pop(0)
            if item["programId"] not in self.indexes:
                self.videos.insert(0, TVItem(item))
                self.indexes.insert(0, item["programId"])

    def find_date(self, content):
        match = self.DATE.search(content)
        if match is not None:
            return match.group(0)

        return False

    def remove_expired_videos(self):
        """Remove from the video data base the expired video.

        """
        lgg.info('Remove expired videos')
        expired = []
        now = datetime.now()
        for idx, item in enumerate(self.videos):
            if item.expire < now:
                expired.append(idx)

        for e in reversed(expired):
            self.videos.pop(e)
            self.indexes.pop(e)

    def get_description(self, item):
        """Returns the long description of a video.

        Args:
        item -- TVItem instance
        """
        if not item.description:
            if item.idx in self.summaries:
                desc, item.origin, item.anno = self.summaries[item.idx]
                if desc is None:
                    self.search_description(item)

                else:
                    item.description = desc.strip()

            else:
                self.search_description(item)

        return item.description

    def save_txt(self, txt):
        with open("itemdata", "wb") as outf:
            jsn = json.dumps(txt, sort_keys=True, indent=4, 
                         separators=(',', ': '), ensure_ascii=False)
            outf.write(jsn.encode('utf-8', 'replace'))

    def search_description(self, item):
        """Fetch the long description of a video.

        If the long description can't be found the short one is used or
        at least the teaser which always exists.

        Args:
        item -- TVItem instance
        """
        desc, origin, anno, duration, rea, country = '', '', '', '', '', ''
        page = self.get_page(item.url)

        if page:
            try:
                txt = page.split("window.__INITIAL_STATE__ =")[1]
                txt = txt.split("</script>")[0]
                txt = txt.split("window.__CLASS_IDS__")[0]
                txt = txt.strip(' ;\n')
                j = json.loads(txt)
                #self.save_txt(j)
            except Exception as why:
                lgg.info("Read error: %s" % why)
                pass

            else:
                try:
                    data = j["pages"]
                    code = data["currentCode"]
                    data = data["list"][code]["zones"][0]["data"][0]
                except Exception as why:
                    lgg.info("No keys 'page-zones-data'")
                    lgg.info("Reason: %s" % why)
                    return
                try:
                    desc = data["fullDescription"]
                    if desc is None:
                        desc = ""
                    desc = desc.replace("&#39;", "'")
                except KeyError:
                    lgg.info("No key 'fullDescription'")
                    desc = ""

                if not desc:
                    try:
                        desc = data["shortDescription"]
                        if desc is None:
                            desc = "No description"
                        desc = desc.replace("&#39;", "'")
                    except KeyError:
                        lgg.info("No key 'shortDescription'")
                        desc = "No description"

                try:
                    duration = self.format_duration(data["duration"])
                except KeyError:
                    pass

                try:
                    credits = data["credits"]
                except KeyError:
                    pass

                else:
                    for credit in credits:
                        if isinstance(credit, dict):
                            if credit["code"] == "REA":
                                try:
                                    rea = ", ".join(credit["values"])
                                except:
                                    rea = "Unknow"

                            elif credit["code"] == "COUNTRY":
                                try:
                                    country = ", ".join(credit["values"])
                                except:
                                    pass

                            elif credit["code"] == "PRODUCTION_YEAR":
                                try:
                                    anno = ", ".join(credit["values"])
                                except:
                                    pass


                origin = "%s, %s" %(rea, country)


        item.description = desc.strip()
        item.origin = origin
        item.anno = anno
        item.duration = duration
        self.summaries[item.idx] = item.description, item.origin, item.anno

    def format_duration(self, value):
        if value is not None:
            min_, sec = divmod(value, 60)
            if min_ < 60:
                return "%s min. %s sec." %(min_, sec)

            hours, min_ = divmod(min_, 60)
            return "%s h. %s min. %s sec." %(hours, min_, sec)

    def check_thumbnails(self):
        """Verify if all thumbnails exists.

        """
        lgg.info('Check thumbnails')
        for idx, index, in enumerate(self.indexes):
            image = os.path.join(self.cache, "%s.jpg" % index)
            image2 = os.path.join(self.cache, "%s.JPG" % index)
            if not (os.path.isfile(image) or os.path.isfile(image2)):
                image = self.videos[idx].get_pixmap()
                if image is None:
                    image = self.load_image(self.videos[idx])
            
            self.videos[idx].pixmap = image
        lgg.info('Thumbnails completed')

    def get_page(self, url):
        """Get a web page.

        Args:
        url -- the url of the page
        """
        lgg.info('Load page: %s' % url)
        try:
            content = urllib.request.urlopen(url).read()
            return str(content.decode('utf-8', 'replace'))
        except Exception as why:
            lgg.info('urllib2 error: %s, %s' % (url, why))
            self.urlliberror = (url, why)
            return False

    def load_image(self, item):
        """Get the image file for the thumbnail.

        Args:
        item -- video item instance
        """
        ext = os.path.splitext(item.pix_url)[1] or ".jpg"
        path = os.path.join(self.cache, "%s%s" % (item.idx, ext))
        for i in range(2):
            try:
                with open(path, 'wb') as objfile:
                    f = urllib.request.urlopen(item.pix_url, timeout=5)
                    objfile.write(f.read())
                return path
            except Exception as why:
                lgg.info("Can't load image: %s" % item.pix_url)
                lgg.info("Reason: %s" % why)

        return "medias/noPreview.png"

    def get_stream_urls(self, item):
        """Get the list of the stream's urls of a video.

        Args:
        item -- TVItem instance
        """ 
        lgg.info(u'Get stream URLs for: {0}'.format(item.url))
        items = item.url.split('/')
        idx = items.index(self.lang)
        base = 'https://api.arte.tv/api/player/v1/config'
        plat = '?platform=ARTEPLUS7'
        lng = '&lang=%s_%s' %(self.lang, self.lang.upper())
        lnk = '%s/%s/%s%s%s&config=arte_tvguide' %(base, self.lang, 
                                                   item.idx, plat, lng)
        links = self.get_page(lnk)
        if links:
            streams, _ = self.read_streams(links)
            if streams:
                item.streams = self.sort_streams(streams)
                return True

            lgg.info("Not found, try other link")
            lnk = '%s/%s/%s?autostart' %(base, self.lang, item.idx)
            links = self.get_page(lnk)
            if links:
                streams, error = self.read_streams(links)
                if streams:
                    item.streams = self.sort_streams(streams)
                    return True

            warn = WarningBox(5, error, self.ui)
            rep = warn.exec_()
            return False

        else:
            warn = WarningBox(4, self.urlliberror, self.ui)
            rep = warn.exec_()
            return False

    def read_streams(self, content):
        """Un-json the list of video's streams.

        Args:
        content -- str(web page)
        """
        try:
            dct = json.loads(content)
            vsr = dct["videoJsonPlayer"]['VSR']
            if vsr:
                return dct["videoJsonPlayer"]['VSR'], None

            else:
                reason = dct["videoJsonPlayer"]["custom_msg"]["msg"]
        except Exception as why:
            lgg.info("Can't read the stream urls: %s" % why)
            reason = "Not available in replay"

        return False, reason

    def sort_streams(self, dct):
        """Clean the video's streams.

        Args:
        dct -- dict(streams)
        """
        keys = list(dct.keys())
        for key in keys:
            if dct[key]["mediaType"] in ("hls", "rtmp"):
                dct.pop(key)

        return dct

    def load_summaries_index(self):
        """Load the video's descriptions list.

        """
        self.summaries = {}
        path = "%s_%s" %(self.core.workspace['tv_summaries'], self.lang)
        if os.path.isfile(path):
            with open(path, 'r') as inf:
                try:
                    self.summaries = json.loads(inf.read())
                except Exception as why:
                    lgg.info("Can't load summaries: %s" % why)

    def save_summaries_index(self):
        """save the video's descriptions list.

        """
        path = "%s_%s" %(self.core.workspace['tv_summaries'], self.lang)
        jsn = json.dumps(self.summaries, indent=4, 
                        separators=(',', ': '), ensure_ascii=False)
        with open(path, 'wb') as outf:
            outf.write(jsn.encode('utf-8', 'replace'))

    def get_site_pages(self):
        """Returns the list of the arte TV Guide web pages.

        """
        if self.lang == 'fr':
            return ARTE_TV

        return [l.replace('/fr/', '/de/').replace('FR', 'DE') for l in ARTE_TV]

    def remove_old_videos(self, value):
        def remove():
            self.clean_thumbnails()
            self.clean_summaries()
        Thread(target=remove).start()

    def clean_thumbnails(self):
        """Remove old thumbnails from the folder 'plusPreviews'.

        """
        lgg.info("Clean thumbnail folder")
        fld = self.core.workspace['tv_tmp']
        thumbs = glob.glob(os.path.join(fld, '*.*'))
        for thumb in thumbs:
            index = os.path.basename(thumb).split('.')[0]
            if not index in self.indexes:
                try:
                    os.remove(thumb)
                except Exception as why:
                    lgg.warning("Can't remove thumbnail: %s" % thumb)
                    lgg.warning("Reason: %s" % why)

    def clean_summaries(self):
        obsoletes = []
        for key in self.summaries.keys():
            if not key in self.indexes:
                obsoletes.append(key)

        for key in obsoletes:
            del self.summaries[key]

    def edit_summary(self, item):
        """Show the summary.

        Args:
        idx -- video item 
        """
        if not item.origin:
            self.search_description(item)

        if item.duration is None:
            item.duration = ''

        desc = self.get_description(item)
        if not desc:
            desc = item.teaser or "No informations"

        self.ui.pitch_editor.clear()
        font = QFont("sans", 10)
        c_form = QTextCharFormat()
        c_form.setFont(font)
        font1 = QFont("sans", 12)
        font1.setWeight(75)
        c_form1 = QTextCharFormat()
        c_form1.setFont(font1)
        self.ui.pitch_editor.setCurrentCharFormat(c_form1)
        self.ui.pitch_editor.appendPlainText(item.title)
        self.ui.pitch_editor.setCurrentCharFormat(c_form)
        t = "  %s %s\n  %s %s\n" % (item.date, item.duration, 
                                    item.origin, item.anno)
        self.ui.pitch_editor.insertPlainText(t)
        if "</" in desc:
            self.ui.pitch_editor.appendHtml("    %s" % desc)

        else:
            self.ui.pitch_editor.appendPlainText("    %s" % desc)

        self.ui.pitch_editor.verticalScrollBar().setValue(0)
        QCoreApplication.processEvents()

    def download_later(self, item):
        lgg.info('Call deferred task with item %s' % item.title)
        if self.get_stream_urls(item):
            item.set_file_name(self.core.cfg['add_date'])
            if not self.core.cfg['tv_quality'] in item.streams:
                self.find_quality(item)

            else:
                item.quality = self.core.cfg['tv_quality']

        self.core.create_deferred_task(item)

    def set_filter(self, key=None):
        if key is None:
            key = self.ui.filter_led.text()

        if not key:
            return

        keys = [k.lower() for k in key.split(' ')]
        selected = []
        for idx, v in enumerate(self.videos):
            title = v.title.lower()
            for k in keys:
                if k in title:
                    selected.append((idx, v))
                    break

        if selected:
            self.ui.arte_list.display_filtered_videos(selected)
            self.ui.showall_btn.setEnabled(True)

        self.update_filter_label(len(selected))
        self.update_filter_keys(key)

    def remove_filter(self):
        self.update_filter_label()
        self.ui.showall_btn.setEnabled(False)
        self.ui.filter_led.blockSignals(True)
        self.ui.filter_led.setText('')
        self.ui.filter_led.setInformativeText(_('Enter a keyword'))
        self.ui.filter_led.blockSignals(False)
        QCoreApplication.processEvents()
        self.ui.arte_list.display_videos()

    def update_filter_label(self, num=None):
        if num is None:
            self.ui.result_lbl.hide()
            return

        self.ui.result_lbl.show()
        if num == 0:
            text = 'No such videos'

        elif num == 1:
            text = 'One video found'

        else:
            text = '{0} videos found'.format(str(num))

        self.ui.result_lbl.setText(text)

    def update_filter_keys(self, new):
        keys = self.core.cfg["filter_keys"]
        if new in keys:
            keys.remove(new)

        keys.insert(0, new)
        self.core.cfg["filter_keys"] = keys[:10]

    def resize_thumbnails(self, size):
        if self.core.cfg['thumb1'] != size:
            self.ui.arte_list.setIconSize(QSize(size, size))
            self.core.cfg['thumb1'] = size

    def start_download(self):
        """Configure and start the downloading.

        """
        if not self.ui.tv_basket.lst_movies:
            return

        video = self.ui.tv_basket.lst_movies[0]
        if video.streams is not None:
            if video.quality is None:
                if self.core.cfg['tv_quality'] is not None:
                    video.quality = self.core.cfg['tv_quality']

                else:
                    ls = [None, 'fr', 'de']
                    video.quality = "HTTP_MP4_EQ_%s" % str(ls.index(self.lang))

            while 1:
                try:
                    stream = video.streams[video.quality]["url"]
                except KeyError:
                    self.configure_downloading(video)

                else:
                    break

            target = self.core.get_target_path(self.core.cfg["videos_folder"])
            self.video_loaded = self.get_video_path(video)
            if not self.initiate_loader(stream, target):
                return

            if not self.core.check_free_space(self.file_size, 
                                              self.core.cfg["videos_folder"]):
                self.reset_loading_widgets()
                return

            self.update_widgets_on_loading()
            lgg.info('Start download %s quality: %s file size: %s'
                        %(video.fname, video.quality, self.file_size))
            lgg.info('URL: %s' % stream)
            self.is_loading = True
            Thread(target=self.loader.download_mp4).start()

    def initiate_loader(self, url, fname):
        self.file_size = self.loader.config_loader(url, fname)
        if isinstance(self.file_size, int):
            return True

        lgg.info('Loading error: %s' % self.file_size)
        return False

    def update_widgets_on_loading(self):
        self.ui.tv_basket.disable_settings()
        self.ui.tv_basket_buttons.on_progress()
        self.ui.action_Download.setEnabled(False)
        self.ui.action_Cancel.setEnabled(True)

    def reset_loading_widgets(self):
        self.ui.tv_basket_buttons.dwnld_btn.setEnabled(True)
        self.ui.tv_basket_buttons.cancel_btn.setEnabled(False)
        self.ui.action_Download.setEnabled(True)
        self.ui.action_Cancel.setEnabled(False)

    def get_video_path(self, item):
        """Return the path/name of the video file.

        Args:
        item -- TVItem instance
        """
        fld = self.core.cfg["videos_folder"]
        path = os.path.join(fld, item.fname + ".mp4")
        count = 1
        while os.path.isfile(path):
            path = os.path.join(fld, item.fname + "(%s).mp4" % count)
            count += 1

        return path

    def on_item_added(self, idx):
        """Verify the stream's urls for a video.

        Called when a video is added into the downloading basket.

        Args:
        idx -- index of the video
        """
        video = self.videos[idx]

        if self.get_stream_urls(video):
            video.set_file_name(self.core.cfg["add_date"])
            if not self.core.cfg['tv_quality'] in video.streams:
                self.find_quality(video)

            else:
                video.quality = self.core.cfg['tv_quality']

        else:
            # Stream urls page not found
            self.ui.tv_basket.remove_item()

    def find_quality(self, item):
        """Select the quality of a video.

        Args:
        item -- TVItem instance
        """
        lang = str((None, 'fr', 'de').index(self.lang))
        qualities = ("EQ_", "HQ_", "SQ_", "MQ_")
        for q in qualities:
            quality = "HTTP_MP4_%s%s" % (q, lang)
            if quality in item.streams:
                item.quality = quality
                return

        # Try with HTTPS
        for q in qualities:
            quality = "HTTPS_%s%s" % (q, lang)
            if quality in item.streams:
                item.quality = quality
                return

        lgg.warning("Can't find mp4 quality for: %s" % item.title)

    def cancel_download(self):
        """Stop the current downloading thread.

        """
        lgg.info('Cancel downloading')
        self.loading_aborted = True
        self.loader.kill_request = True

    def display_progress(self, percent, speed, remain):
        """Show the downloading progress informations.

        Args:
        percent -- percent downloaded
        speed -- bandwidth in Ko/sec.
        remain -- estimated remaining time
        """
        rtime = int(remain / speed)
        speed = self.locale.toString(int(speed / 1024))
        self.ui.tv_basket_buttons.progressBar.setValue(percent)
        self.ui.tv_basket_buttons.progress_lbl.setText(
                                "Speed: %s Kio/sec." % speed)
        m, s = divmod(rtime, 60)
        self.ui.tv_basket_buttons.progress_lbl2.setText(
                                "Remaining time: %s min. %s sec." %(m, s))

    def on_downloading_finished(self, fname):
        """Rename the video downloaded.

        Args:
        fname -- temporary file name
        """
        lgg.info('File %s downloaded' % fname)
        self.is_loading = False
        item = self.ui.tv_basket.remove_item(0)
        self.ui.tv_basket_buttons.clear()
        self.ui.action_Cancel.setEnabled(False)
        if not fname:
            self.handle_downloading_error()
            return

        try:
            os.rename(fname, self.video_loaded)
            lgg.info('File renamed: %s' % os.path.basename(self.video_loaded))
        except Exception as why:
            lgg.warning("Can't rename %s" % self.video_loaded)
            lgg.warning("Reason: %s" % why)

        self.check_file_size()
        self.copy_summary(item, self.video_loaded)
        self.copy_thumbnail(item, self.video_loaded)
        if self.ui.tv_basket.lst_movies:
            self.start_download()

    def check_file_size(self):
        if self.loading_aborted:
            # The downloading was stopped by the user
            self.loading_aborted = False
            return

        expected = self.loader.size
        loaded = os.path.getsize(self.video_loaded)
        if loaded != expected:
            # TODO format numbers
            lgg.warning("The downloading seems incomplete!")
            lgg.warning("Size expected: %s downloaded: %s" %(expected, loaded))
            fname = os.path.basename(self.video_loaded)
            warn = WarningBox(9, (fname, expected, loaded), self.ui)
            rep = warn.exec_()

    def configure_downloading(self, video):
        """Call the downloading config dialog box.

        Args:
        video -- TVItem instance
        """
        lgg.info('Configure downloading of %s' % video.title)
        self.cfg = TvDownloadingConfig(self)
        self.cfg.configure(video.streams, video, 
                           self.core.cfg['tv_quality'],
                           self.core.cfg['lang'], 
                           date=self.core.cfg['add_date'])
        self.cfg.exec_()

    def set_preferred_quality(self, quality):
        self.core.cfg['tv_quality'] = quality

    def handle_downloading_error(self):
        """Display a warning box when the downloading failed.

        """
        url = self.loader.url
        error = self.loader.last_error
        lgg.warning("Downloading error: %s" % error)
        lgg.warning("URL: %s" % url)
        warn = WarningBox(8, (url, error), self.ui)
        rep = warn.exec_()
        if self.ui.tv_basket.lst_movies:
            self.start_download()

    def on_empty_list(self):
        self.ui.tv_basket_buttons.dwnld_btn.setEnabled(False)
        self.ui.action_Download.setEnabled(False)
        self.ui.action_Cancel.setEnabled(False)

    def copy_summary(self, video, fname):
        """Save the summary into a file.

        Args:
        video -- TVItem instance
        fname -- path-name of the video downloaded
        """
        if not self.core.cfg['pitch_plus']:
            return

        if self.core.cfg['pitch_plus_unique']:
            path = os.path.join(os.path.dirname(fname), 'index')

        else:
            path = os.path.splitext(fname)[0]

        title = video.fname
        date, length, author, anno = "", "", "", ""
        try:
            date = video.date
        except:
            pass
        try:
            length = video.duration
        except:
            pass
        try:
            author = video.origin
        except:
            pass
        try:
            anno = video.anno
        except:
            pass
        summary = self.get_description(video)

        txt = "\n* %s, (%s) %s\n  %s, %s\n\n%s" %(title, date, length, author, 
                                                  anno, summary)
        try:
            with open(path, 'a') as outf:
                outf.write(txt)
                lgg.info("Summary saved")
        except Exception as why:
            lgg.warning("Can't write summary: %s" % path)
            lgg.warning("Reason: %s" % why)

    def copy_thumbnail(self, video, fname):
        """Save the thumbnail into a file.

        Args:
        video -- TVItem instance
        fname -- path-name of the video downloaded
        """
        if self.core.cfg['copythumb_plus']:
            path = os.path.splitext(fname)[0] + ".jpg"
            orig = video.pixmap
            shutil.copy(orig, path)
            lgg.info("Thumbnail saved")

class Parser:
    def __init__(self, url):
        self.url = url
        self.videos = []
        self.indexes = []
        diff = timedelta(days=10)
        last = date.today() - diff
        now = last.strftime("%y-%m-%d")
        self.day_before = self.get_day_before(last)
        self.user_agent = 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:49.0)'\
                  ' Gecko/20100101 Firefox/49.0'
        url = url.replace("DAY", now)
        items = self.get_json(self.fetch_page(url))
        if items:
            for item in items:
                video = self.clean_item(item)
                if video:
                    self.videos.insert(0, video)
                    self.indexes.insert(0, video['id'])


    def get_day_before(self, start):
        now = start
        diff = timedelta(days=1)
        while 1:
            now = now + diff
            yield now.strftime("%y-%m-%d")

    def fetch_page(self, url):
        lgg.info('Fetch page: %s' % url)
        req = urllib.request.Request(url, data=None, 
                                 headers={"User-Agent": self.user_agent})
        try:
            content = urllib.request.urlopen(req)
        except Exception as why:
            lgg.warning('urllib error: %s, %s' % (url, why))
            return False
        return str(content.read().decode('utf-8', 'replace'))

    def get_next_page(self):
        url = self.url.replace("DAY", next(self.day_before))
        items = self.get_json(self.fetch_page(url))
        videos = []
        indexes = []
        if items:
            for item in items:
                video = self.clean_item(item)
                if video:
                    videos.append(video)
                    indexes.append(video['id'])

        return videos, indexes

    def get_json(self, content):
        if content:
            try:
                data = json.loads(content)
                for z in data['zones']:
                    if z["code"] == "listing_TV_GUIDE":
                        return z['data']
            except Exception as why:
                lgg.warning("Read json error: %s" % why)
        lgg.warning("listing_TV_GUIDE NOT FOUND")
        return False

    def clean_item(self, item):
        video = {}
        rejected = True
        try:
            video["url"] = item["url"]
            video['id'] = item["programId"]
        except KeyError:
            # No url, no video
            lgg.warning("No URL or ID")
            return False

        try:
            for i in item["stickers"]:
                if "label" in i and i["label"] == "PLAYABLE":
                    rejected = False
        except KeyError:
            # No playable
            pass

        if rejected:
            return False

        video["subtitle"] = item.get("subtitle", "")
        video["headerText"] = item.get("headerText", "")
        video["title"] = item.get("title", "No title")
        video["description"] = item.get("fullDescription", "")
        video["duration"] = item.get("duration", "")
        video["teaser"] = item.get("teaserText", "")
        video["programId"] = video['id']
        video['adult'] = None
        video['views'] = 0
        video["rights_end"] = item.get("videoRightsEnd", "")
        bd = item.get("broadcastDates", "")
        if isinstance(bd, list):
            bd = bd[0]
        video['scheduled_on'] = bd
        img = False
        img1 = False
        try:
            if "landscape" in item["images"]:
                img = item["images"]["landscape"]["resolutions"][0]["url"]

            elif "square" in item["images"]:
                img1 = item["images"]["square"]["resolutions"][0]["url"]

            video["thumbnail_url"] = img or img1
        except Exception as why:
                lgg.warning("No thumbnail: %s" % why)
                video["thumbnail_url"] = None
        return video


class TVItem:
    """Define a video item.

    """
    def __init__(self, attr, lang='fr'):
        self.idx = attr['id']
        self.title = attr['title']
        # Very short description
        self.teaser = attr.get('teaser', '')
        self.sched = attr['scheduled_on']
        self.expire = self.format_expiration_date(attr['rights_end'])
        self.date = attr['scheduled_on']
        self.duration = self.format_duration(attr['duration'])
        self.url = attr['url']
        self.subtitle = attr['subtitle']
        self.adult = attr['adult']
        self.pix_url = attr['thumbnail_url']
        self.pixmap = None
        # description, origin and anno must be extracted from the video page
        self.description = attr.get('description', '')
        self.header = attr.get('headerText', '')
        self.origin = ''
        self.anno = ''
        self.views = attr['views']
        self.streams = None
        self.quality = None
        self.lang = lang
        self.set_file_name(False)
        if self.date is not None:
            if not 'T' in self.date:
                self.date = self.format_date(attr['scheduled_on'], 
                                             attr['rights_end'])
            else:
                # Remove the time zone '+02'
                self.date = self.date.split('+')[0].replace('T', ' ').lstrip("'")
        
    def format_date(self, begin, end):
        """Return the release date and time of the video.

        """
        # The release value contents only the date, we take the expiration
        # time and we add two hours
        if not end:
            end = begin
        end = end.split('T')[1].rstrip('Z')
        h, m, s = end.split(':')
        h = str((int(h) + 2) % 24)
        return "%s %s" %(begin, ':'.join([h, m, s]))

    def format_expiration_date(self, date):
        # Expiration date is '2016-04-21T03:15:00Z'
        try:
            return datetime.strptime(date, "%Y-%m-%dT%H:%M:%SZ")
        except ValueError:
            return datetime.now()

    def format_duration(self, value):
        if value is not None:
            min_, sec = divmod(value, 60)
            if min_ < 60:
                return "%s min. %s sec." %(min_, sec)

            hours, min_ = divmod(min_, 60)
            return "%s h. %s min. %s sec." %(hours, min_, sec)

    def get_pixmap(self):
        title = self.title.lower()
        if 'xenius' in title:
            return 'medias/Xenius.jpg'

        if 'journal' in title:
            if 'best' in title:
                return 'medias/best_journal.jpg'
            if 'junior' in title:
                return 'medias/journal_junior.jpg'
            return 'medias/journal.jpg'

        if 'dessous des cartes' in title or 'mit offenen karten' in title:
            return 'medias/dessous_cartes.jpg'

    def set_file_name(self, date):
        f = self.title.replace('/', '-').replace('»', '_').replace('«', '_')
        self.fname = f.replace('\\', ' ')
        if date and self.date is not None:
            self.fname += '_%s' % self.date.replace(' ', '_')

